import sys
import re
import logging
import io
import pysam
from .vcf_reader import VcfReader
from .dbsnp_filter import dbSnpFilter, clinvar_path_annot
from .gnomad_filter import GnomadFilter
from .vcf_filter import VcfFilter
from .vep_filter import VepFilter
from .snpeff_filter import SnpEffFilter
from .cadd_filter import CaddFilter
from .sample_filter import SampleFilter
from .ped_file import PedFile, Individual, PedError
from .family_filter import FamilyFilter, ControlFilter
from .family_filter import RecessiveFilter, DominantFilter, DeNovoFilter
from .burden_counter import BurdenCounter
from .var_by_region import VarByRegion
from .region_iter import RegionIter
from .gt_annotator import GtAnnotator
from .spliceai_filter import SpliceAiFilter, filter_on_splice_ai
from .info_filter import InfoFilter
from .g2p import G2P


class VaseRunner(object):

    def __init__(self, args):
        self._twirler = ['-', '\\', '|', '/']
        self.args = args
        self._set_logger()
        if self.args.debug:
            self.logger.debug(vars(args))
        self.input = VcfReader(self.args.input, logger=self.logger)
        self.var_stream = self.input
        self.keep_filters = None
        self.exclude_filters = None
        if args.keep_filters:
            self.keep_filters = set(args.keep_filters)
        if args.exclude_filters:
            self.exclude_filters = set(args.exclude_filters)
        self.var_types = self._parse_var_type_arg()
        self.info_filter = self._parse_info_filters()
        self.prev_cadd_phred = False
        self.prev_cadd_raw = False
        self.prev_splice_ai = False
        self._get_prev_annotations()
        self.vcf_filters = self.get_vcf_filter_classes()
        self.cadd_filter = self.get_cadd_filter()
        self.splice_ai_filter = self.get_splice_ai_filter()
        self.gt_annotators = self.get_gt_annotators()
        self.ped = None
        if args.ped:
            self.ped = PedFile(args.ped)
        self.g2p = None
        self.gene_filter = None
        self.retrieving_by_region = False
        self.gt_args = dict(
            gq=args.gq,
            dp=args.dp,
            max_dp=args.max_dp,
            het_ab=args.het_ab,
            hom_ab=args.hom_ab,
            min_control_dp=args.control_dp,
            max_control_dp=args.control_max_dp,
            min_control_gq=args.control_gq,
            control_het_ab=args.control_het_ab,
            control_hom_ab=args.control_hom_ab,
            con_ref_ab=args.control_max_ref_ab,
            sv_gq=args.sv_gq,
            sv_dp=args.sv_dp,
            sv_max_dp=args.sv_max_dp,
            sv_het_ab=args.sv_het_ab,
            sv_hom_ab=args.sv_hom_ab,
            sv_min_control_dp=args.sv_control_dp,
            sv_max_control_dp=args.sv_control_max_dp,
            sv_min_control_gq=args.sv_control_gq,
            sv_control_het_ab=args.sv_control_het_ab,
            sv_control_hom_ab=args.sv_control_hom_ab,
            sv_con_ref_ab=args.sv_control_max_ref_ab,
            del_dhffc=args.duphold_del_dhffc,
            dup_dhbfc=args.duphold_dup_dhbfc,
            control_del_dhffc=args.control_duphold_del_dhffc,
            control_dup_dhbfc=args.control_duphold_dup_dhbfc,
        )
        # region/bed/gene_bed args are mutually exclusive (handled by parser)
        if args.region is not None:
            self.var_stream = VarByRegion(self.input,
                                          region_iter=RegionIter(args.region),
                                          stream=args.stream,
                                          exclude=args.exclude_regions)
            self.retrieving_by_region = True
        if args.bed is not None:
            self.logger.info("Reading, sorting and merging intervals in " +
                             "{}".format(args.bed))
            self.var_stream = VarByRegion(self.input,
                                          bed=args.bed,
                                          stream=args.stream,
                                          exclude=args.exclude_regions)
            self.retrieving_by_region = True
            self.logger.info("Finished processing intervals.")
        if args.gene_bed is not None:
            self.logger.info("Reading, sorting and merging intervals in " +
                             "{}".format(args.gene_bed))
            self.gene_filter = VarByRegion(self.input,
                                           bed=args.gene_bed,
                                           gene_targets=True,
                                           snpeff_mode=args.snpeff,
                                           stream=args.stream,
                                           exclude=args.exclude_regions)
            if self.args.csq is None:
                self.args.csq = ['all']
            if self.args.biotypes is None:
                self.args.biotypes = ['all']
            self.var_stream = self.gene_filter
            self.retrieving_by_region = True
            self.logger.info("Finished processing intervals.")
        if args.g2p is not None:
            self.g2p = G2P(args.g2p, snpeff_mode=args.snpeff)
        self.csq_filter, self.post_spliceai_csq_filter = self.get_csq_filters()
        self.sample_filter = None
        self.burden_counter = None
        if args.burden_counts:
            if args.n_cases or args.n_controls:
                self.logger.warning("--n_cases and --n_controls arguments are " +
                                 "ignored when using --burden_counter.")
            self.burden_counter = BurdenCounter(
                self.input,
                args.burden_counts,
                is_gnomad=args.gnomad_burden,
                cases=args.cases,
                controls=args.controls,
                gq=args.gq,
                dp=args.dp,
                max_dp=args.max_dp,
                het_ab=args.het_ab,
                hom_ab=args.hom_ab,
            )
        elif args.cases or args.controls:
            self.sample_filter = SampleFilter(
                self.input,
                cases=args.cases,
                controls=args.controls,
                n_cases=args.n_cases,
                n_controls=args.n_controls,
                confirm_missing=args.confirm_control_gts,
                **self.gt_args)
        self.de_novo_filter = None
        self.dominant_filter = None
        self.recessive_filter = None
        self.family_filter = None
        self.control_filter = None
        self.variant_cache = VariantCache(snpeff_mode=args.snpeff)
        self.use_cache = False
        self.prog_interval = args.prog_interval
        self.log_progress = args.log_progress
        self.strict_recessive_inheritance = args.strict_recessive
        self.report_fhs = self.get_report_filehandles()
        seg_info = list()
        if args.de_novo:
            seg_info.extend(self._get_de_novo_filter())
        if args.biallelic or args.singleton_recessive:
            seg_info.extend(self._get_recessive_filter())
        if args.dominant or args.singleton_dominant:
            seg_info.extend(self._get_dominant_filter())
        if args.check_g2p_inheritance:
            if (not self.dominant_filter and not self.recessive_filter
                    and not self.de_novo_filter):
                raise ValueError("--check_g2p_inheritance option requires " +
                                   "at least one segregation option (e.g. " +
                                   "--recessive, --dominant or --de_novo) " +
                                   "to be selected.")
        self._check_got_inherit_filter()
        self._set_seg_annot_cleanup(seg_info)
        if self.args.output is None:
            self.args.output = '-'
        self.add_vase_header()
        self.out = pysam.VariantFile(self.args.output,
                                     mode='w',
                                     header=self.input.header.header)
        self.var_count = 0
        self.var_written = 0
        self.var_filtered = 0
        self.prog_string = ''
        self.prog_updates = 0

    def update_progress(self, record):
        if (self.args.no_progress or self.var_count % self.prog_interval):
            return
        s = ('{:,} variants processed, '.format(self.var_count) +
             '{:,} filtered, '.format(self.var_filtered) +
             '{:,} written...'.format(self.var_written) +
             ' at pos {}:{}'.format(record.chrom, record.pos))
        if self.retrieving_by_region and self.var_stream.region_iter:
            s += " (processing region {}/{})".format(
                self.var_stream.region_iter.current_index + 1,
                len(self.var_stream.region_iter.intervals))
        if self.log_progress:
            self.logger.info(s)
        else:
            twirl = self._twirler[self.prog_updates % 4]
            s = '\r' + s + ' ' + twirl
            if len(self.prog_string) > len(s):
                sys.stderr.write('\r' + ' ' * len(self.prog_string))
            sys.stderr.write(s)
        self.prog_string = s
        self.prog_updates += 1

    def run(self):
        ''' Run VCF filtering/annotation using args from bin/vase'''
        self.logger.info('Starting variant processing')
        with self.var_stream:
            for vase_record in self.var_stream:
                self.process_record(vase_record)
                self.var_count += 1
                self.update_progress(vase_record.record)
        self.finish_up()
        if (self.prog_string and not self.log_progress and
                not self.args.no_progress):
            sys.stderr.write('\r' + '-' * len(self.prog_string) + '\n')
        self.logger.info('Finished processing {:,} {}.'.format(
            self.var_count,
            self._var_or_vars(self.var_count)))
        self.logger.info('{:,} {} filtered.'.format(
            self.var_filtered,
            self._var_or_vars(self.var_filtered)))
        self.logger.info('{:,} {} written.'.format(
            self.var_written,
            self._var_or_vars(self.var_written)))

    def process_record(self, vase_record):
        if self.filter_global(vase_record):
            self.var_filtered += 1
            return
        filter_alleles = None
        if self.var_types:
            filter_alleles = [
                x.var_type not in self.var_types
                for x in vase_record.DECOMPOSED_ALLELES
            ]
            if all(filter_alleles):
                # no ALT matches any variant type asked for
                self.var_filtered += 1
                return
        filter_alleles, filter_csq = self.filter_alleles_external(
            vase_record, filter_alleles)
        if all(filter_alleles):
            # all alleles should be filtered
            self.var_filtered += 1
            return
        if self.sample_filter:
            for i in range(1, len(vase_record.record.alleles)):
                r = self.sample_filter.filter(vase_record, i)
                if r:
                    filter_alleles[i - 1] = True
                if all(filter_alleles):
                    # all alleles should be filtered
                    self.var_filtered += 1
                    return
        dom_filter_alleles = list(filter_alleles)
        if self.control_filter:
            for i in range(1, len(vase_record.record.alleles)):
                if dom_filter_alleles[i - 1]:  # no need to filter again
                    continue
                r = self.control_filter.filter(vase_record, i)
                if r:
                    dom_filter_alleles[i - 1] = True
        self.remove_previous_inheritance_filters(vase_record.record)
        denovo_hit = False
        dom_hit = False
        recessive_hit = False
        if self.dominant_filter:
            dom_hit = self.dominant_filter.process_record(
                vase_record, dom_filter_alleles, filter_csq)
        if self.de_novo_filter:
            denovo_hit = self.de_novo_filter.process_record(
                vase_record, dom_filter_alleles, filter_csq)
        if self.recessive_filter:
            recessive_hit = self.recessive_filter.process_record(
                vase_record, filter_alleles, filter_csq)
        if self.use_cache:
            if denovo_hit or dom_hit or recessive_hit:
                keep_record_anyway = False
                if self.args.min_families < 2:
                    keep_record_anyway = denovo_hit or dom_hit
                self.variant_cache.add_record(vase_record, keep_record_anyway)
            else:
                self.variant_cache.check_record(vase_record)
                self.var_filtered += 1
            if self.variant_cache.output_ready:
                self.output_cache()
        elif self.de_novo_filter or self.dominant_filter:
            if denovo_hit or dom_hit:
                self.output_record(vase_record)
                if self.burden_counter:
                    # getting relevant alleles and feats is a bit of a fudge
                    # using annotations added by dom/denovo filter
                    b_filt_al = [True] * len(vase_record.record.alts)
                    b_filt_csq = [[True] * len(vase_record.CSQ)] * len(
                        b_filt_al)
                    if dom_hit:
                        b_filt_al, b_filt_csq = self._seg_alleles_from_record(
                            vase_record,
                            self.dominant_filter.prefix,
                            b_filt_al,
                            b_filt_csq)
                    if denovo_hit:
                        b_filt_al, b_filt_csq = self._seg_alleles_from_record(
                            vase_record,
                            self.de_novo_filter.prefix,
                            b_filt_al,
                            b_filt_csq)
                    for i in range(len(b_filt_al)):
                        if not b_filt_al[i]:
                            feat = (vase_record.CSQ[j]['Feature']
                                    for j in range(len(vase_record.CSQ))
                                    if not b_filt_csq[i][j])
                            self.burden_counter.count_samples(
                                vase_record, feat, i, 1)
            else:
                self.var_filtered += 1
        else:
            if all(filter_alleles):
                self.var_filtered += 1
                return
            if self.burden_counter:
                self.burden_counter.count(vase_record, filter_alleles,
                                          filter_csq)
            self.output_record(vase_record)

    def output_record(self, vase_record):
        for gt_anno in self.gt_annotators:
            gt_anno.annotate(vase_record)
        self.out.write(vase_record.record)
        self.var_written += 1

    def output_cache(self, final=False):
        keep_ids = set()
        burden_vars = dict()  # dict of inheritance model to segregants
        if self.recessive_filter:
            vid_to_seg = self.recessive_filter.process_potential_recessives(
                final=final)
            keep_ids.update(vid_to_seg.keys())
            if self.burden_counter:
                burden_vars['recessive'] = vid_to_seg
        if self.dominant_filter and self.args.min_families > 1:
            vid_to_seg = self.dominant_filter.process_dominants(final=final)
            keep_ids.update(vid_to_seg.keys())
            if self.burden_counter:
                burden_vars['dominant'] = vid_to_seg
        if self.de_novo_filter and self.args.min_families > 1:
            vid_to_seg = self.de_novo_filter.process_de_novos(final=final)
            keep_ids.update(vid_to_seg.keys())
            if self.burden_counter:
                burden_vars['de_novo'] = vid_to_seg
        if final:
            self.variant_cache.add_cache_to_output_ready()
        for var in self.variant_cache.output_ready:
            if var.can_output or var.var_id in keep_ids:
                self.output_record(var.record)
            else:
                self.var_filtered += 1
        if self.burden_counter:
            self._burden_from_cache(burden_vars)
        self.variant_cache.output_ready = []

    def _burden_from_cache(self, model_to_vars):
        for model in ('dominant', 'de_novo'):
            if model in model_to_vars:
                for v in model_to_vars[model]:
                    for seg in model_to_vars[model][v]:
                        self.burden_counter.count_samples(
                            seg.record, seg.features, seg.allele - 1, 1)
        # we count recessives last as the max per allele (2) is higher than for
        # dom/de novos (otherwise max per sample could get lowered to 1)
        if 'recessive' in model_to_vars:
            for v in model_to_vars['recessive']:
                for seg in model_to_vars['recessive'][v]:
                    self.burden_counter.count_samples(seg.record, seg.features,
                                                      seg.allele - 1, 2)

    def finish_up(self):
        if self.use_cache:
            self.output_cache(final=True)
        self.out.close()
        if self.burden_counter:
            self.burden_counter.output_counts()
        for fh in self.report_fhs.values():
            if fh is not None:
                fh.close()

    def filter_alleles_external(self, vase_record, remove_alleles=None):
        '''
            Return True or False for each allele indicating whether an
            allele should be filtered based on information from VEP,
            dbSNP, ClinVar or gnomAD.
        '''
        # remove_alleles indicates whether allele should be filtered;
        # keep_alleles indicates whether allele should be kept, overriding any
        # indications in remove_alleles (e.g. if labelled pathogenic in
        # ClinVar)
        # remove_csq indicates for each VEP CSQ whether that CSQ should be
        # ignored
        if not remove_alleles:
            remove_alleles = [False] * len(vase_record.record.alts)
        keep_alleles = [False] * len(vase_record.record.alts)
        matched_alleles = [False] * len(vase_record.record.alts)
        remove_csq = None
        # if allele is '*' should be set to filtered
        for i in range(1, len(vase_record.record.alleles)):
            if vase_record.record.alleles[i] == '*':
                remove_alleles[i - 1] = True
        # filter on provided INFO field filters
        if self.info_filter:
            r_alts = self.info_filter.filter(vase_record.record)
            self._set_to_true_if_true(remove_alleles, r_alts)
            if all(remove_alleles):
                # bail out now if no valid allele and not keeping clinvar
                return remove_alleles, remove_csq
        # check VCF's internal AF
        if self.args.af or self.args.min_af:
            r_alts = self.filter_on_af(vase_record.record)
            self._set_to_true_if_true(remove_alleles, r_alts)
            if (not self.args.clinvar_path and all(remove_alleles)):
                # bail out now if no valid allele and not keeping clinvar
                # path variants - if using clinvar path we have to ensure we
                # haven't got a path variant with a non-qualifying allele
                return remove_alleles, remove_csq
        # check VCF's internal AC
        if self.args.ac or self.args.min_ac:
            r_alts = self.filter_on_ac(vase_record.record)
            self._set_to_true_if_true(remove_alleles, r_alts)
            if (not self.args.clinvar_path and all(remove_alleles)):
                # bail out now if no valid allele and not keeping clinvar
                return remove_alleles, remove_csq
        # check functional consequences
        splice_alleles = None
        if self.csq_filter:
            r_alts, remove_csq = self.csq_filter.filter(vase_record)
            self._set_to_true_if_true(remove_alleles, r_alts)
            if self.prev_splice_ai and (self.args.splice_ai_min_delta
                                        or self.args.splice_ai_max_delta):
                splice_alleles, splice_csq = filter_on_splice_ai(
                    vase_record,
                    min_delta=self.args.splice_ai_min_delta,
                    max_delta=self.args.splice_ai_max_delta,
                    snpeff_mode=self.args.snpeff,
                    check_symbol=True,
                    canonical_csq=self.args.canonical)
                self._set_to_false_if_true(remove_alleles, splice_alleles)
                self._set_to_false_if_true(remove_csq, splice_csq)
            elif self.splice_ai_filter:
                splice_alleles, splice_csq = (
                    self.splice_ai_filter.annotate_or_filter(
                        vase_record,
                        check_symbol=True,
                        canonical_csq=self.args.canonical,
                        snpeff_mode=self.args.snpeff))
                if (self.args.splice_ai_min_delta
                        or self.args.splice_ai_max_delta):
                    # RETAIN Alleles/csq if SpliceAI scores meet threshold
                    self._set_to_false_if_true(remove_alleles, splice_alleles)
                    self._set_to_false_if_true(remove_csq, splice_csq)
            if self.post_spliceai_csq_filter:
                # filter with VEP annots again in case of AF/biotype failures
                r_alts, r_csq = self.post_spliceai_csq_filter.filter(
                    vase_record)
                self._set_to_true_if_true(remove_alleles, r_alts)
                self._set_to_true_if_true(remove_csq, r_csq)
            if (not self.args.clinvar_path and all(remove_alleles)):
                # bail out now if no valid consequence
                return remove_alleles, remove_csq
        elif (self.splice_ai_filter or self.args.splice_ai_min_delta
              or self.args.splice_ai_max_delta):
            # if not filtering with CSQ annotations filter if spliceAI
            # thresholds not met
            if self.prev_splice_ai and (self.args.splice_ai_min_delta
                                        or self.args.splice_ai_max_delta):
                splice_alleles, splice_csq = filter_on_splice_ai(
                    vase_record,
                    min_delta=self.args.splice_ai_min_delta,
                    max_delta=self.args.splice_ai_max_delta,
                    snpeff_mode=self.args.snpeff,
                    check_symbol=True,
                    canonical_csq=self.args.canonical)
                remove_alleles = [not(x) or y for x, y in zip(splice_alleles,
                                                              remove_alleles)]
            else:
                splice_alleles, splice_csq = (
                    self.splice_ai_filter.annotate_or_filter(vase_record))
                if (self.args.splice_ai_min_delta
                        or self.args.splice_ai_max_delta):
                    remove_alleles = [not(x) or y for x, y in
                                      zip(splice_alleles, remove_alleles)]
        cadd_drop = [False] * len(vase_record.record.alts)
        if self.prev_cadd_phred and self.args.cadd_phred:
            r_alts = self.filter_on_existing_cadd_phred(vase_record)
            self._set_to_true_if_true(cadd_drop, r_alts)
        if self.prev_cadd_raw and self.args.cadd_raw:
            r_alts = self.filter_on_existing_cadd_raw(vase_record)
            self._set_to_true_if_true(cadd_drop, r_alts)
        if self.cadd_filter:
            r_alts = self.cadd_filter.annotate_or_filter(vase_record)
            self._set_to_true_if_true(cadd_drop, r_alts)
        if any(cadd_drop):
            if splice_alleles is not None and not self.args.prioritise_cadd:
                # splice_alleles (i.e. over SpliceAI threshold) override
                # filtering because CADD < threshold
                cadd_drop = [x and not y for x, y in zip(cadd_drop,
                                                         splice_alleles)]
            self._set_to_true_if_true(remove_alleles, cadd_drop)
        if (not self.args.clinvar_path and all(remove_alleles)):
            # bail out now if no valid consequence and not keeping clinvar
            # path variants - if using clinvar path we have to ensure we
            # haven't got a path variant with a non-qualifying consequence
            return remove_alleles, remove_csq
        for f in self.vcf_filters:
            r, k, m = f.annotate_and_filter_record(vase_record)
            # should only overwrite value of remove_alleles[i] or
            # keep_alleles[i] with True, not False (e.g. if already set to
            # be filtered because of a freq in ExAC we shouldn't set to
            # False just because it is absent from dbSNP)
            self._set_to_true_if_true(remove_alleles, r)
            self._set_to_true_if_true(matched_alleles, m)
            self._set_to_true_if_true(keep_alleles, k)
        if self.prev_freqs:
            r, m = self.filter_on_existing_freq(vase_record.record)
            self._set_to_true_if_true(remove_alleles, r)
            self._set_to_true_if_true(matched_alleles, m)
        if self.prev_homs:
            r, m = self.filter_on_existing_homs(vase_record.record)
            self._set_to_true_if_true(remove_alleles, r)
            self._set_to_true_if_true(matched_alleles, m)
        if self.prev_builds:
            r, m = self.filter_on_existing_build(vase_record.record)
            self._set_to_true_if_true(remove_alleles, r)
            self._set_to_true_if_true(matched_alleles, m)
        if self.prev_clinvar:
            k, m = self.filter_on_existing_clnsig(vase_record.record)
            self._set_to_true_if_true(matched_alleles, m)
            self._set_to_true_if_true(keep_alleles, k)
        verdict = []
        for i in range(len(remove_alleles)):
            if keep_alleles[i]:
                verdict.append(False)
            elif remove_alleles[i]:
                verdict.append(True)
            elif self.args.filter_known and matched_alleles[i]:
                verdict.append(True)
            elif self.vcf_filters:
                # only apply filter_novel if we have supplied an external vcf
                if self.args.filter_novel and not matched_alleles[i]:
                    verdict.append(True)
                else:
                    verdict.append(False)
            else:
                verdict.append(False)
        return verdict, remove_csq

    def filter_on_af(self, record):
        remove = [False] * len(record.alts)
        if self.args.filtering_an and self.an_below_threshold(record):
            return remove
        try:
            af = record.info['AF']
            for i in range(len(remove)):
                if self.args.af:
                    if af[i] is not None and af[i] > self.args.af:
                        remove[i] = True
                if self.args.min_af:
                    if af[i] is None or af[i] < self.args.min_af:
                        remove[i] = True
        except KeyError:
            self.logger.debug("No 'AF' in INFO at {}:{}".format(
                record.chrom, record.pos))
            if 'AN' not in record.info:
                self.logger.warning("No 'AF' or 'AN' in INFO at {}:{}".format(
                    record.chrom, record.pos) + " - will not filter on AF")
            elif 'AC' not in record.info:
                self.logger.warning("No 'AF' or 'AC' in INFO at {}:{}".format(
                    record.chrom, record.pos) + " - will not filter on AF")
            else:
                self.logger.debug("Trying AC/AN instead")
                return self.filter_on_ac_over_an(record)
        return remove

    def an_below_threshold(self, record):
        try:
            return (record.info['AN'] < self.args.filtering_an)
        except KeyError:
            self.logger.warning("No 'AN' in INFO at {}:{}".format(record.chrom,
                                                               record.pos))
            return True

    def an_under_minimum(self, record):
        try:
            return (record.info['AN'] < self.args.min_an)
        except KeyError:
            self.logger.warning("No 'AN' in INFO at {}:{}".format(record.chrom,
                                                               record.pos))
            return True

    def filter_on_ac_over_an(self, record):
        remove = [False] * len(record.alts)
        try:
            ac = record.info['AC']
            an = record.info['AN']
            af = [ac[i] / an if an > 0 else 0 for i in range(len(ac))]
            for i in range(len(remove)):
                if self.args.af:
                    if af[i] is not None and af[i] > self.args.af:
                        remove[i] = True
                if self.args.min_af:
                    if af[i] is None or af[i] < self.args.min_af:
                        remove[i] = True
        except KeyError:
            self.logger.warning("Missing 'AN' or 'AC' field in INFO at {}:{}".
                             format(record.chrom, record.pos) + " - will not"
                             " filter on AF")
        return remove

    def filter_on_ac(self, record):
        remove = [False] * len(record.alts)
        try:
            ac = record.info['AC']
            for i in range(len(remove)):
                if self.args.ac:
                    if ac[i] is not None and ac[i] > self.args.ac:
                        remove[i] = True
                if self.args.min_ac:
                    if ac[i] is None or ac[i] < self.args.min_ac:
                        remove[i] = True
        except KeyError:
            self.logger.warning("No 'AC' in INFO at {}:{}".format(
                record.chrom, record.pos))
        return remove

    def filter_on_existing_cadd_phred(self, record):
        remove = [False] * len(record.alts)
        if 'CADD_PHRED_score' in record.info:
            phreds = record.info['CADD_PHRED_score']
            for i in range(len(remove)):
                if (phreds[i] is not None and
                        phreds[i] < self.args.cadd_phred):
                    remove[i] = True
        return remove

    def filter_on_existing_cadd_raw(self, record):
        remove = [False] * len(record.alts)
        if 'CADD_raw_score' in record.info:
            raws = record.info['CADD_raw_score']
            for i in range(len(remove)):
                if (raws[i] is not None and raws[i] < self.args.cadd_raw):
                    remove[i] = True
        return remove

    def filter_on_existing_freq(self, record):
        remove = [False] * len(record.alts)
        matched = [False] * len(record.alts)
        for annot in self.prev_freqs:
            if annot not in record.info or record.info[annot] is None:
                continue
            for i in range(len(remove)):
                if record.info[annot][i] is not None:
                    matched[i] = True
                    if self.args.freq:
                        if record.info[annot][i] >= self.args.freq:
                            remove[i] = True
                    if self.args.min_freq:
                        if record.info[annot][i] < self.args.min_freq:
                            remove[i] = True
        return remove, matched

    def filter_on_existing_homs(self, record):
        remove = [False] * len(record.alts)
        matched = [False] * len(record.alts)
        for annot in self.prev_homs:
            if annot not in record.info or record.info[annot] is None:
                continue
            for i in range(len(remove)):
                if record.info[annot][i] is not None:
                    matched[i] = True
                    n = record.info[annot][i]
                    if n > self.args.max_gnomad_homozygotes:
                        remove[i] = True
        return remove, matched

    def filter_on_existing_build(self, record):
        remove = [False] * len(record.alts)
        matched = [False] * len(record.alts)
        for annot in [x for x in self.prev_builds if x in record.info]:
            if annot not in record.info or record.info[annot] is None:
                continue
            for i in range(len(remove)):
                if record.info[annot][i] is not None:
                    matched[i] = True
                    if self.args.build:
                        if record.info[annot][i] <= self.args.build:
                            remove[i] = True
                    if self.args.max_build:
                        if record.info[annot][i] > self.args.max_build:
                            remove[i] = True
        return remove, matched

    def filter_on_existing_clnsig(self, record):
        keep = [False] * len(record.alts)
        matched = [False] * len(record.alts)
        for annot in self.prev_clinvar:
            if annot not in record.info or record.info[annot] is None:
                continue
            for i in range(len(keep)):
                if record.info[annot][i] is not None:
                    matched[i] = True
                    if self.args.clinvar_path:
                        if any(x for x in clinvar_path_annot
                               if x in record.info[annot][i].split('|')):
                            keep[i] = True
        return keep, matched

    def filter_global(self, record):
        ''' Return True if record fails any global variant filters.'''
        if record.alts is None:
            return True
        if self.args.pass_filters:
            if record.filter.keys() != ['PASS']:
                return True
        elif self.keep_filters:  # --pass and --keep_filters are mutually excl.
            if not self.keep_filters.issuperset(record.filter):
                return True
        if self.exclude_filters:
            if self.exclude_filters.intersection(record.filter):
                return True
        if self.args.variant_quality is not None:
            if record.qual < self.args.variant_quality:
                return True
        if self.args.max_alt_alleles is not None:
            if (len([x for x in record.alleles if x != '*']) >
                    self.args.max_alt_alleles + 1):
                return True
        if self.args.min_an and self.an_under_minimum(record):
            return True
        if self.args.filter_asterisk_only_calls:
            if len(record.alleles) == 2 and record.alleles[1] == '*':
                return True
        return False

    def get_csq_filters(self):
        if self.args.csq is None and self.args.impact is None \
           and self.args.g2p is None:
            return None, None
        elif self.args.snpeff:
            return self.get_snpeff_filters()
        return self.get_vep_filters()

    def get_snpeff_filters(self):
        for option in ('missense_filters', 'splice_filters', 'loftee',
                       'vep_af', 'pathogenic'):
            if getattr(self.args, option):
                raise ValueError("--{} option is incompatible".format(option) +
                                 " with --snpeff option. This option " +
                                 "requires filtering using VEP annotations.")
        csq_filter = SnpEffFilter(
            vcf=self.input,
            csq=self.args.csq,
            impact=self.args.impact,
            biotypes=self.args.biotypes,
            retain_labels=self.args.retain_labels,
            filter_flagged_features=self.args.flagged_features,
            gene_filter=self.gene_filter,
            blacklist=self.args.feature_blacklist,
            g2p=self.g2p,
            check_g2p_consequence=self.args.check_g2p_consequence,
            logging_level=self.logger.level)
        if self._need_post_spliceai_csq_filter():
            post_splice_ai_filter = SnpEffFilter(
                vcf=self.input,
                csq=['all'],
                biotypes=self.args.biotypes,
                gene_filter=self.gene_filter,
                blacklist=self.args.feature_blacklist,
                logging_level=self.logger.level)
        else:
            post_splice_ai_filter = None
        return csq_filter, post_splice_ai_filter

    def get_vep_filters(self):
        if self.args.no_vep_freq:
            if self.args.vep_af:
                self.logger.warning("Ignoring --vep_af argument because " +
                                 "--no_vep_af argument is in use.")
            vep_freq = None
            vep_min_freq = None
            vep_af = []
        else:
            vep_freq = self.args.freq
            vep_min_freq = self.args.min_freq
            vep_af = self.args.vep_af
        csq_filter = VepFilter(
            vcf=self.input,
            csq=self.args.csq,
            impact=self.args.impact,
            canonical=self.args.canonical,
            biotypes=self.args.biotypes,
            in_silico=self.args.missense_filters,
            filter_unpredicted=self.args.filter_unpredicted,
            keep_any_damaging=self.args.keep_if_any_damaging,
            loftee=self.args.loftee,
            splice_in_silico=self.args.splice_filters,
            splice_filter_unpredicted=self.args.splice_filter_unpredicted,
            splice_keep_any_damaging=self.args.splice_keep_if_any_damaging,
            retain_labels=self.args.retain_labels,
            filter_flagged_features=self.args.flagged_features,
            freq=vep_freq,
            min_freq=vep_min_freq,
            filter_known=self.args.filter_known,
            filter_novel=self.args.filter_novel,
            afs=vep_af,
            gene_filter=self.gene_filter,
            blacklist=self.args.feature_blacklist,
            pathogenic=self.args.pathogenic,
            no_conflicted=self.args.no_conflicted,
            g2p=self.g2p,
            check_g2p_consequence=self.args.check_g2p_consequence,
            logging_level=self.logger.level)
        if self._need_post_spliceai_csq_filter():
            # workaround - need to VepFilter in 2 stages if using SpliceAI
            post_spliceai_csq_filter = VepFilter(
                vcf=self.input,
                csq=['all'],
                canonical=self.args.canonical,
                biotypes=self.args.biotypes,
                freq=vep_freq,
                min_freq=vep_min_freq,
                filter_known=self.args.filter_known,
                filter_novel=self.args.filter_novel,
                afs=vep_af,
                gene_filter=self.gene_filter,
                blacklist=self.args.feature_blacklist,
                logging_level=self.logger.level)
        else:
            post_spliceai_csq_filter = None
        return csq_filter, post_spliceai_csq_filter

    def _need_post_spliceai_csq_filter(self):
        if self.splice_ai_filter:
            return True
        return self.prev_splice_ai and (self.args.splice_ai_min_delta or
                                        self.args.splice_ai_max_delta)

    def get_cadd_filter(self):
        if self.args.cadd_directory or self.args.cadd_files:
            cadd_args = {
                'cadd_files': self.args.cadd_files,
                'cadd_dir': self.args.cadd_directory,
                'min_phred': self.args.cadd_phred,
                'min_raw_score': self.args.cadd_raw,
                'to_score': self.args.missing_cadd_scores
            }
            cf = CaddFilter(**cadd_args)
            for f, d in cf.info_fields.items():
                self.logger.debug("Adding {} annotation")
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
            return cf
        else:
            if self.args.cadd_phred is not None and not self.prev_cadd_phred:
                raise ValueError("--cadd_phred cutoff given but VCF does " +
                                 "not have CADD_PHRED_score INFO field " +
                                 "in its header and no CADD score files " +
                                 "have been specified with --cadd_files or" +
                                 " --cadd_dir arguments.")
            if self.args.cadd_raw is not None and not self.prev_cadd_raw:
                raise ValueError("--cadd_raw cutoff given but VCF does " +
                                 "not have CADD_raw_score INFO field " +
                                 "in its header and no CADD score files " +
                                 "have been specified with --cadd_files " +
                                 "or --cadd_dir arguments.")
        return None

    def get_splice_ai_filter(self):
        if self.args.splice_ai_vcfs:
            splice_ai_args = {
                'vcfs': self.args.splice_ai_vcfs,
                'min_delta': self.args.splice_ai_min_delta,
                'max_delta': self.args.splice_ai_max_delta,
                'to_score': self.args.missing_splice_ai_scores,
                'logging_level': self.logger.level,
            }
            sf = SpliceAiFilter(**splice_ai_args)
            for f, d in sf.info_fields.items():
                self.logger.debug("Adding {} annotation")
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
            return sf
        else:
            if not self.prev_splice_ai:
                if self.args.splice_ai_min_delta is not None:
                    raise ValueError(
                        "--splice_ai_min_delta cutoff given " +
                        "but VCF does not have SpliceAI INFO " +
                        "field in its header and no SpliceAI " +
                        "score files have been specified with " +
                        "--splice_ai_vcfs argument.")
                if self.args.splice_ai_max_delta is not None:
                    raise ValueError(
                        "--splice_ai_max_delta cutoff given " +
                        "but VCF does not have SpliceAI INFO " +
                        "field in its header and no SpliceAI " +
                        "score files have been specified with " +
                        "--splice_ai_vcfs argument.")
        return None

    def get_vcf_filter_classes(self):
        filters = []
        uni_args = {'logger': self.logger}
        if self.args.freq is not None:
            uni_args["freq"] = self.args.freq
        if self.args.min_freq is not None:
            uni_args["min_freq"] = self.args.min_freq
        # get dbSNP filter
        for dbsnp in self.args.dbsnp:
            prefix = self.check_info_prefix('VASE_dbSNP')
            kwargs = {
                "vcf": dbsnp,
                "prefix": prefix,
                "clinvar_path": self.args.clinvar_path,
            }
            if self.args.build is not None:
                kwargs['build'] = self.args.build
            if self.args.max_build is not None:
                kwargs['max_build'] = self.args.max_build
            kwargs.update(uni_args)
            dbsnp_filter = dbSnpFilter(**kwargs)
            filters.append(dbsnp_filter)
            for f, d in dbsnp_filter.added_info.items():
                self.logger.debug("Adding dbSNP annotation {}".format(f))
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
        # get gnomAD/ExAC filters
        for gnomad in self.args.gnomad:
            prefix = self.check_info_prefix('VASE_gnomAD')
            kwargs = {
                "vcf": gnomad,
                "prefix": prefix,
                "pops": self.args.gnomad_pops,
                "max_homozygotes": self.args.max_gnomad_homozygotes
            }
            kwargs.update(uni_args)
            gnomad_filter = GnomadFilter(**kwargs)
            filters.append(gnomad_filter)
            for f, d in gnomad_filter.added_info.items():
                self.logger.debug("Adding gnomAD/ExAC annotation {}".format(f))
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
        # get other VCF filters
        for var_filter in self.args.vcf_filter:
            vcf_and_id = var_filter.split(',')
            if len(vcf_and_id) < 2:
                raise ValueError(
                    "Expected two or more comma separated " +
                    "values for --vcf_filter argument '{}'".format(var_filter))
            prefix = self.check_info_prefix('VASE_' + vcf_and_id[1])
            kwargs = {
                "vcf": vcf_and_id[0],
                "prefix": prefix,
                "annotations": vcf_and_id[2:]
            }
            kwargs.update(uni_args)
            vcf_filter = VcfFilter(**kwargs)
            filters.append(vcf_filter)
            for f, d in vcf_filter.added_info.items():
                self.logger.debug("Adding {} annotation from {}".format(
                    f, vcf_and_id[0]))
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
        return filters

    def get_gt_annotators(self):
        gt_annos = []
        if self.args.dng_vcf:
            for vcf in self.args.dng_vcf:
                g = (GtAnnotator(vcf, ['PP_DNM', 'PP_NULL']))
                for f, d in g.header_fields.items():
                    self.input.header.add_header_field(name=f,
                                                       dictionary=d,
                                                       field_type='FORMAT')
                gt_annos.append(g)
        return gt_annos

    def _seg_alleles_from_record(self, record, prefix, filter_al,
                                 filter_csq):
        ps = prefix + '_samples'
        pf = prefix + '_features'
        f_al = [
            False if record.info[ps][i] != '.' else filter_al[i]
            for i in range(len(filter_al))
        ]
        f_csq = [[
            False if (record.info[pf][i] is not None
                      and record.CSQ[j]['Feature'] in
                      record.info[pf][i].split('|'))
            else filter_csq[i][j] for j in range(len(record.CSQ))
        ] for i in range(len(filter_al))]
        return f_al, f_csq

    def _get_prev_annotations(self):
        self.prev_annots = set()
        self.info_prefixes = set()
        for info in self.input.header.info:
            match = re.search(r'^(VASE_\w+)_\w+(_\d+)?', info)
            if match:
                self.prev_annots.add(info)
                self.info_prefixes.add(match.group(1))
                self.logger.debug("Identified previously annotated VASE INFO" +
                                  " field '{}'".format(info))
        self._parse_prev_vcf_filter_annotations()
        if 'CADD_PHRED_score' in self.input.header.info:
            inf = self.input.header.info['CADD_PHRED_score']
            if (inf.number == 'A' and inf.type == 'Float'):
                self.prev_cadd_phred = True
        if 'CADD_raw_score' in self.input.header.info:
            inf = self.input.header.info['CADD_raw_score']
            if (inf.number == 'A' and inf.type == 'Float'):
                self.prev_cadd_raw = True
        if 'SpliceAI' in self.input.header.info:
            if self.input.header.info['SpliceAI'].number == '.':
                self.prev_splice_ai = True

    def _parse_prev_vcf_filter_annotations(self):
        frq_annots = []
        hom_annots = []
        bld_annots = []
        cln_annots = []
        get_matching = self.args.filter_known or self.args.filter_novel
        if not self.args.ignore_existing_annotations:
            if self.args.freq or self.args.min_freq or get_matching:
                for annot in sorted(self.prev_annots):
                    match = re.search(
                        r'^VASE_dbSNP|gnomAD(_\d+)?_(CAF|AF)(_\w+)?', annot)
                    if match:
                        inf = self.input.header.info[annot]
                        if (inf.number == 'A' and inf.type == 'Float'):
                            self.logger.info(
                                "Found previous allele frequency " +
                                "annotation '{}'".format(annot))
                            if len(match.groups()
                                   ) == 3 and match.group(3) is not None:
                                pop = match.group(3).replace("_", "")
                                if pop.upper() not in [
                                        x.upper()
                                        for x in self.args.gnomad_pops
                                ]:
                                    self.logger.info(
                                        "Ignoring {} ".format(annot) +
                                        "annotation as not in " +
                                        "populations specified by " +
                                        "--gnomad_pops")
                                    continue
                            frq_annots.append(annot)
            if self.args.max_gnomad_homozygotes is not None:
                for annot in sorted(self.prev_annots):
                    match = re.search(
                        r'^VASE_gnomAD(_\d+)?_(Hom|Hemi|nhomalt)(_\w+)?',
                        annot)
                    if match:
                        inf = self.input.header.info[annot]
                        if (inf.number == 'A' and inf.type == 'Integer'):
                            self.logger.info("Found previous Hom/Hemi " +
                                             "annotation '{}'".format(annot))
                            if (len(match.groups()) == 3 and
                                    match.group(3) is not None):
                                pop = match.group(3).replace("_", "")
                                if pop.upper() not in [
                                        x.upper()
                                        for x in self.args.gnomad_pops
                                ]:
                                    self.logger.info(
                                        "Ignoring {} ".format(annot) +
                                        "annotation as not in " +
                                        "populations specified by " +
                                        "--gnomad_pops")
                                    continue
                            hom_annots.append(annot)
            if self.args.build or self.args.max_build or get_matching:
                for annot in sorted(self.prev_annots):
                    match = re.search(r'^VASE_dbSNP(_\d+)?_dbSNPBuildID',
                                      annot)
                    if match:
                        inf = self.input.header.info[annot]
                        if inf.number == 'A' and inf.type == 'Integer':
                            self.logger.info("Found previous dbSNP build " +
                                             "annotation '{}'".format(annot))
                            bld_annots.append(annot)
            if self.args.clinvar_path or get_matching:
                for annot in sorted(self.prev_annots):
                    match = re.search(r'^VASE_dbSNP(_\d+)?_CLNSIG', annot)
                    if match:
                        inf = self.input.header.info[annot]
                        if inf.number == 'A' and inf.type == 'String':
                            self.logger.info("Found previous ClinVar " +
                                             "annotation '{}'".format(annot))
                            cln_annots.append(annot)
            # if using gnomAD file for burden counting use internal annotations
            # for frequency filtering, if specified
            if self.args.gnomad_burden and (self.args.freq
                                            or self.args.min_freq):
                # check only relatively outbred pops by default
                pops = set(
                    ("POPMAX", "AFR", "AMR", "EAS", "FIN", "NFE", "SAS"))
                for annot in self.input.header.info:
                    match = re.search(r'''^AF_([A-Z]+)$''', annot)
                    if match and match.group(1).upper() in pops:
                        inf = self.input.header.info[annot]
                        if inf.number == 'A' and inf.type == 'Float':
                            self.logger.info("Found gnomAD allele frequency " +
                                             "annotation '{}'".format(annot))
                            frq_annots.append(annot)
        self.prev_freqs = tuple(frq_annots)
        self.prev_homs = tuple(hom_annots)
        self.prev_builds = tuple(bld_annots)
        self.prev_clinvar = tuple(cln_annots)

    def check_info_prefix(self, name):
        if name in self.info_prefixes:
            self.logger.debug(
                "INFO field {} already exists - trying another".format(name))
            match = re.search(r'_(\d+)$', name)
            if match:
                # already has an appended '_#' - increment and try again
                i = int(match.group(1))
                name = re.sub(match.group(1) + '$', str(i + 1), name)
                return self.check_info_prefix(name)
            else:
                # append _1
                return self.check_info_prefix(name + '_1')
        self.info_prefixes.add(name)
        return name

    def add_vase_header(self):
        '''
            Write a VCF header for output that consists of the input VCF
            header data plus program arguments and any new INFO/FORMAT
            fields.
        '''

        vase_opts = []
        for k, v in vars(self.args).items():
            vase_opts.append('--{} {}'.format(k, v))
        self.input.header.header.add_meta(key="vase",
                                          value=str.join(" ", vase_opts))

    def get_report_filehandles(self):
        fhs = {
            'recessive': None,
            'dominant': None,
            'de_novo': None,
        }
        if self.args.report_prefix is not None:
            if self.args.biallelic or self.args.singleton_recessive:
                f = self.args.report_prefix + ".recessive.report.tsv"
                fhs['recessive'] = open(f, 'w')
            if self.args.dominant or self.args.singleton_dominant:
                f = self.args.report_prefix + ".dominant.report.tsv"
                fhs['dominant'] = open(f, 'w')
            if self.args.de_novo:
                f = self.args.report_prefix + ".de_novo.report.tsv"
                fhs['de_novo'] = open(f, 'w')
        return fhs

    def _set_to_true_if_true(self, alist, values):
        for i in range(len(alist)):
            if values[i]:
                alist[i] = True

    def _set_to_false_if_true(self, alist, values):
        for i in range(len(alist)):
            if values[i]:
                alist[i] = False

    def _get_family_filter(self):
        if self.family_filter is not None:
            return self.family_filter
        infer = True
        no_ped = False
        if not self.ped:
            if (not self.args.singleton_recessive
                    and not self.args.singleton_dominant):
                raise ValueError("Inheritance filtering options require a " +
                                 "PED file specified using --ped or else " +
                                 "samples specified using " +
                                 "--singleton_recessive or " +
                                 "--singleton_dominant arguments")
            else:
                self.ped = self._make_ped_io()
                no_ped = True
                infer = False
        if self.args.check_g2p_inheritance:
            g2p = self.g2p
        else:
            g2p = None
        for s in self.args.seg_controls:
            indv = Individual(s, s, "0", "0", 0, 1)
            try:
                self.ped.add_individual(indv)
            except PedError:
                pass
        if not no_ped:
            for s in set(self.args.singleton_recessive +
                         self.args.singleton_dominant):
                indv = Individual(s, s, "0", "0", 0, 2)
                try:
                    self.ped.add_individual(indv)
                except PedError:
                    raise ValueError(
                        "Sample '{}' ".format(s) + "specified" +
                        " as either --singleton_recessive or " +
                        "--singleton_dominant already exists " +
                        "in PED file {}".format(self.ped.filename))
        self.family_filter = FamilyFilter(
            ped=self.ped,
            vcf=self.input,
            infer_inheritance=infer,
            g2p=g2p,
            check_g2p_consequence=self.args.check_g2p_consequence,
            logging_level=self.logger.level)
        for s in set(self.args.singleton_dominant):
            self.family_filter.inheritance_patterns[s].append('dominant')
        for s in set(self.args.singleton_recessive):
            self.family_filter.inheritance_patterns[s].append('recessive')

    def _parse_info_filters(self):
        if not self.args.info_filters:
            return None
        ifilters = []
        for expression in self.args.info_filters:
            exp = expression.split()
            if len(exp) != 3:
                raise ValueError(
                    "--info_filters must consist of three quoted" +
                    " values separated by whitespace - for " +
                    "example: 'QD > 4' The provided expression " +
                    " '{}' is invalid.".format(expression))
            ifilters.append(exp)
        return InfoFilter(vcf=self.input, filters=ifilters)

    def _parse_var_type_arg(self):
        if not self.args.var_types:
            return None
        valid = {'SNV', 'INSERTION', 'DELETION', 'MNV', 'SV'}
        v_types = []
        for vt in self.args.var_types:
            if vt == 'INDEL':
                v_types.extend(['INSERTION', 'DELETION'])
            elif vt.upper() in valid:
                v_types.append(vt.upper())
            else:
                raise ValueError("Invalid --var_type provided: {}".format(vt))
        return v_types

    def _make_ped_io(self):
        p_string = ''
        for s in set(self.args.singleton_recessive +
                     self.args.singleton_dominant):
            p_string += str.join("\t", (s, s, "0", "0", "0", "2")) + "\n"
        ped = io.StringIO(p_string)
        return PedFile(ped)

    def _set_logger(self):
        self.logger = logging.getLogger("VASE")
        if self.args.silent:
            self.args.no_warnings = True
            self.args.no_progress = True
        if self.args.debug:
            self.logger.setLevel(logging.DEBUG)
        elif self.args.no_warnings:
            self.logger.setLevel(logging.ERROR)
        elif self.args.quiet:
            self.logger.setLevel(logging.WARNING)
        else:
            self.logger.setLevel(logging.INFO)
        formatter = logging.Formatter(
            '[%(asctime)s] %(name)s - %(levelname)s - %(message)s')
        ch = logging.StreamHandler()
        ch.setLevel(self.logger.level)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

    def _check_got_inherit_filter(self):
        if self.args.de_novo or self.args.biallelic or self.args.dominant:
            if (not self.recessive_filter and not self.de_novo_filter
                    and not self.dominant_filter):
                raise ValueError("No inheritance filters could be created " +
                                   "with current settings. Please check your" +
                                   " ped/sample inputs or run without " +
                                   "--biallelic/--dominant/--de_novo options.")

    def _set_seg_annot_cleanup(self, seg_info):
        for_removal = [x for x in seg_info if x in self.prev_annots]
        if for_removal:
            self.info_to_remove = for_removal
            self.remove_previous_inheritance_filters = self._clean_info
            self.logger.warning(
                "Replacing {:,} previous".format(len(for_removal)) +
                " segregation INFO fields.")
            for f in for_removal:
                self.logger.warning(
                    "Previous {} annotations will be removed".format(f))
        else:
            self.remove_previous_inheritance_filters = lambda *args: None

    def _clean_info(self, record):
        ''' Remove predefined INFO fields from records. '''
        for f in (x for x in self.info_to_remove if x in record.info):
            del record.info[f]

    def _get_dominant_filter(self):
        self._get_family_filter()
        self._get_control_filter()
        self.dominant_filter = DominantFilter(
            self.family_filter,
            self.gt_args,
            min_families=self.args.min_families,
            snpeff_mode=self.args.snpeff,
            report_file=self.report_fhs['dominant'])
        added_info = list(self.dominant_filter.get_header_fields().keys())
        if not self.dominant_filter.affected:
            msg = ("No samples fit a dominant model - can not use dominant " +
                   "filtering")
            if not self.args.biallelic and not self.args.de_novo:
                raise ValueError("Error: " + msg)
            else:
                self.logger.warning(msg + ". Will continue with other models.")
                self.dominant_filter = None
        else:
            for f, d in self.dominant_filter.get_header_fields().items():
                self.logger.debug(
                    "Adding DominantFilter annotation {}".format(f))
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
                if self.args.min_families > 1:
                    self.use_cache = True
        return added_info  # so we know which fields to remove if necessary

    def _get_de_novo_filter(self):
        self._get_family_filter()
        self._get_control_filter()
        self.de_novo_filter = DeNovoFilter(
            self.family_filter,
            self.gt_args,
            min_families=self.args.min_families,
            snpeff_mode=self.args.snpeff,
            report_file=self.report_fhs['de_novo'])
        added_info = list(self.de_novo_filter.get_header_fields().keys())
        if not self.de_novo_filter.affected:
            msg = ("No samples fit a de novo model - can not use de novo " +
                   "filtering")
            if not self.args.biallelic and not self.args.dominant:
                raise ValueError("Error: " + msg)
            else:
                self.logger.warning(msg + ". Will continue with other models.")
                self.de_novo_filter = None
        else:
            for f, d in self.de_novo_filter.get_header_fields().items():
                self.logger.debug(
                    "Adding DeNovoFilter annotation {}".format(f))
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
                if self.args.min_families > 1:
                    self.use_cache = True
        return added_info  # so we know which fields to remove if necessary

    def _get_recessive_filter(self):
        self._get_family_filter()
        self.recessive_filter = RecessiveFilter(
            self.family_filter,
            self.gt_args,
            min_families=self.args.min_families,
            strict=self.strict_recessive_inheritance,
            snpeff_mode=self.args.snpeff,
            report_file=self.report_fhs['recessive'])
        added_info = list(self.recessive_filter.get_header_fields().keys())
        if not self.recessive_filter.affected:
            msg = ("No samples fit a recessive model - can not use biallelic" +
                   " filtering")
            if not self.args.de_novo and not self.args.dominant:
                raise ValueError("Error: " + msg)
            else:
                self.logger.warning(msg + ". Will continue with other models.")
                self.recessive_filter = None
        else:
            self.use_cache = True
            for f, d in self.recessive_filter.get_header_fields().items():
                self.logger.debug(
                    "Adding RecessiveFilter annotation {}".format(f))
                self.input.header.add_header_field(name=f,
                                                   dictionary=d,
                                                   field_type='INFO')
        return added_info  # so we know which fields to remove if necessary

    def _get_control_filter(self):
        if self.control_filter:
            return
        self.control_filter = ControlFilter(vcf=self.input,
                                            family_filter=self.family_filter,
                                            n_controls=self.args.n_controls,
                                            gt_args=self.gt_args)

    def _var_or_vars(self, n):
        if n == 1:
            return 'variant'
        return 'variants'


class VariantCache(object):
    '''
        Store a collection of CachedVariants that can be outputted
        once certain conditions are met (e.g. once we've checked
        whether variants match compound heterozygous variation in a
        gene). Keeps track of VEP Features encountered while adding to
        the cache so that the cache can be released for output once the
        current record is outside of the relevant features.
    '''

    __slots__ = ['cache', 'features', 'output_ready', 'features_from_record']

    def __init__(self, snpeff_mode=False):
        self.cache = []
        self.features = set()
        self.output_ready = []
        if snpeff_mode:
            self.features_from_record = self._get_snpeff_annotation_features
        else:
            self.features_from_record = self._get_vep_annotation_features

    def check_record(self, record):
        '''
            Check whether features in given record are disjoint with
            those in cache and if so move variants from cache to
            output_ready. The given record is NOT added to the cache.
        '''
        these_feats = self.features_from_record(record)
        if self.features and these_feats.isdisjoint(self.features):
            self.add_cache_to_output_ready()
            self.features.clear()

    def add_record(self, record, can_output=False):
        these_feats = self.features_from_record(record)
        if self.features and these_feats.isdisjoint(self.features):
            self.add_cache_to_output_ready()
            self.features = these_feats
        else:
            self.features.update(these_feats)
        self.cache.append(CachedVariant(record, can_output))

    def add_cache_to_output_ready(self):
        ''' Adds items in cache to output_ready and clears cache.'''
        self.output_ready.extend(self.cache)
        self.cache = []

    def _get_snpeff_annotation_features(self, record):
        return set([x['Feature_ID'] for x in record.ANN])

    def _get_vep_annotation_features(self, record):
        return set([x['Feature'] for x in record.CSQ])


class CachedVariant(object):
    '''
        Store a variant that will or might need outputting later
        depending on, for example, determining whether they fit a
        recessive inheritance pattern.
    '''

    __slots__ = ['record', 'can_output', 'var_id']

    def __init__(self, record, can_output=False):
        self.record = record
        self.can_output = can_output
        self.var_id = "{}:{}-{}/{}".format(record.chrom, record.pos,
                                           record.ref, record.alt)
